import numpy as np
import pandas as pd


class Game:
    game_name = 'Unnamed Game'
    players = 0

    def __init__(self, seed=0):
        self.history = []

    def get_state_and_actions(self):  # Does not mutate data!
        return None

    def execute_actions(self, actions):
        return None


class Agent:
    agent_type = 'Unnamed Agent'

    def __init__(self):
        pass

    def get_move(self, sna):
        if not sna:
            return None
        return None

    def end(self, score):
        pass

    def __repr__(self):
        return "{}:{}".format(self.agent_type, self.id)


def run_single_game(game, agents_w_kwargs):
    agents = [agent(**y) for agent, y in agents_w_kwargs]
    game = game()
    game_over = False
    while not game_over:
        snas = game.get_state_and_actions()
        acts = tuple([a.get_move(sna) for a, sna in zip(agents, snas)])
        result = game.execute_actions(acts)
        if result:
            game_over = True
    for res, agent in zip(result, agents):
        agent.end(res)
    return result, game.history


def random_unlimited_tournament(game, agents, n=11):
    agents = {str(a(**k)): (a, k) for a, k in agents}
    agent_names = list(agents.keys())
    results = []
    histories = []
    for i in range(n):
        player_ids = np.random.choice(agent_names, game.players, replace=False)
        result, moves = run_single_game(game, [agents[pid] for pid in player_ids])
        results.append({pid: score for pid, score in zip(player_ids, result)})
        histories.append(moves)
    return pd.DataFrame(results), histories



from ..base import Game


class TicTacToe(Game):
    game_name = 'Tic-Tac-Toe'
    players = 2

    def __init__(self, seed=0):
        self.board = ['']*9
        self.current_player = 0
        self.history = []

    def get_state_and_actions(self):  # Does not mutate data!
        actions = [i for i, s in enumerate(self.board) if s == '']
        sna = {'state': (self.board, self.current_player), 'actions': actions}
        return tuple([sna if i == self.current_player else None for i in range(self.players)])

    def execute_actions(self, actions):  # Does change state
        self.history.append(actions)
        move_to_exec = actions[self.current_player]
        self.board[move_to_exec] = self.current_player
        # check if win
        if self.check_player_win(self.current_player):
            # we have a winner
            return tuple([1.0 if i == self.current_player else 0.0 for i in range(self.players)])
        # check if tie
        if len([x for x in self.board if x == '']) == 0:
            return 0.5, 0.5
        # finalize
        self.current_player = (self.current_player+1) % 2
        return None

    def check_player_win(self, player):  # does not change state
        for j in range(3):
            win = True
            for i in range(3):
                win &= self.board[3*j+i] == player
            if win:
                return win
        for j in range(3):
            win = True
            for i in range(3):
                win &= self.board[j+i*3] == player
            if win:
                return win
        if self.board[0] == self.board[4] == self.board[8] == player:
            return True
        if self.board[2] == self.board[4] == self.board[6] == player:
            return True
        return False



from ..base import Agent


class HumanAgent(Agent):
    agent_type = 'Human Agent'

    def __init__(self, name='basic'):
        self.name = name

    def get_move(self, sna):
        if not sna:
            return None
        state = sna['state']
        actions = sna['actions']
        print(state)
        print(actions)
        while True:
            try:
                choice = int(input("Select an action:"))
                assert choice in actions
            except Exception:
                pass
            else:
                break
        return choice

    def end(self, score):
        pass

    def __repr__(self):
        return self.agent_type + ":" + self.name



from copy import deepcopy

import numpy as np

from ..base import Agent


class SillyAgent(Agent):
    agent_name = 'Silly Agent'

    def __init__(self, name='basic'):
        self.name = name

    def get_move(self, sna):
        if not sna:
            return None
        state = sna['state']
        actions = sna['actions']
        choice = self.make_choice(state, actions)
        return choice

    def end(self, score):
        pass

    def check_player_win(self, state, player):
        for j in range(3):
            win = True
            for i in range(3):
                win &= state[3*j+i] == player
            if win:
                return win
        for j in range(3):
            win = True
            for i in range(3):
                win &= state[j+i*3] == player
            if win:
                return win
        if state[0] == state[4] == state[8] == player:
            return True
        if state[2] == state[4] == state[6] == player:
            return True
        return False

    def make_choice(self, state, actions):
        temp = deepcopy(state[0])
        for action in actions:
            temp[action] = 1.0
            if self.check_player_win(temp, 1.0):
                return action
            else:
                temp = deepcopy(state[0])
        for action in actions:
            temp[action] = 0.0
            if self.check_player_win(temp, 0.0):
                return action
            else:
                temp = deepcopy(state[0])
        return np.random.choice(actions)



from ..base import random_unlimited_tournament
from ..agents import RandomAgent
from ..games import TicTacToe


def test_random_unlimited_tournament():
    results = random_unlimited_tournament(TicTacToe, ((RandomAgent, {'name': 'p1'}), (RandomAgent, {'name': 'p2'})))
    assert(len(results) == 2)



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='yomi',
    version='0.0.1',
    description='Tools for running tournaments of AI agents against each-other',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bluecoconut/yomi',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=['numpy', 'pandas'],
    extras_require={  # Optional
        'dev': ['pytest', 'flake8'],
    }
)



import numpy as np

from ..base import Agent


class RandomAgent(Agent):
    agent_type = 'Random Agent'

    def __init__(self, name='basic'):
        self.name = name

    def get_move(self, sna):
        if not sna:
            return None
        # state = sna['state']
        actions = sna['actions']
        choice = np.random.choice(actions)
        return choice

    def end(self, score):
        pass

    def __repr__(self):
        return self.agent_type + ":" + self.name
